@@ -170,13 +170,9 @@ int ident_url_relatif(const char *lien, const char *origin_adr,
 #endif
 #if HTS_USEOPENSSL
   } else if (strfield(lien, "https://")) {
-    if (SSL_is_available) {
-      // Note: ftp:foobar.gif is not valid
-      if (ident_url_absolute(lien, adr, fil) == -1) {
-        ok = -1;                // erreur URL
-      }
-    } else {
-      ok = -1;
+    // Note: ftp:foobar.gif is not valid
+    if (ident_url_absolute(lien, adr, fil) == -1) {
+      ok = -1;                // erreur URL
     }
 #endif
   } else if ((scheme) && ((!strfield(lien, "http:"))
@@ -489,7 +485,7 @@ void longfile_to_83(int mode, char *n83, char *save) {
      /:;?\#*~
      0x00-0x1f and 0x80-0xff
    */
-  for(i = 0; save[i] != 0; i++) {
+  for(i = 0, j = 0; save[i] != 0; i++) {
     char a = save[i];
 
     if (a >= 'a' && a <= 'z') {
@@ -498,10 +494,14 @@ void longfile_to_83(int mode, char *n83, char *save) {
       if (!
           ((a >= 'A' && a <= 'Z') || (a >= '0' && a <= '9') || a == '_'
            || a == '.')) {
+      if (j != 0 && save[j - 1] == '_') {
+        continue;  // avoid __
+      }
       a = '_';
     }
-    save[i] = a;
+    save[j++] = a;
   }
+  save[j] = '\0';
 
   i = j = 0;
   while((i < max) && (save[j]) && (save[j] != '.')) {
